/* Copyright David Leonard, 2000. All rights reserved. */
/* $Id: eval.c,v 1.4 2000/05/22 00:59:32 leonard Exp $ */

#include <stdio.h>
#include <math.h>
#include "node.h"

typedef struct value {
	union {
		scalar_t	scalar;
		vector_t	vector;
		node_t		node;
	} u;
} value_t;

scalar_t	eval_index(node_t, vector_t, scalar_t);
scalar_t	eval_sum(node_t, vector_t);
vector_t	eval_vector(node_t, sym_t);
vector_t	gen_vector(node_t, sym_t);
vector_t	gen_range(node_t, sym_t);
sym_t		enter_name(ident_t, sym_t, node_t);

void
eval_error(n, msg)
	node_t n;
	const char *msg;
{
	int pos = n->pos;
	show_error(pos, msg);
}

void
show_error(pos, msg)
	int pos;
	const char *msg;
{
	extern const char *argv1;
	const char *c;

	if (pos != -1) {
		fprintf(stderr, "%s\n", argv1);
		for (c = argv1; *c; c++) {
			if (pos-- == 0)
				break;
			if (*c == '\t') fprintf(stderr, "\t");
			else	       fprintf(stderr, " ");
		}
		fprintf(stderr, "^\n");
	}
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

/* Try to evaluate an expression in a scalar context */
void
eval(n, sym, vp)
	node_t n;
	sym_t sym;
	value_t *vp;
{
	value_t v1, v2;

	node_t n2;
	vector_t v;
	scalar_t s;
	sym_t y;

	switch (n->type) {
	case NODETYPE_ADD:	
	case NODETYPE_SUB:
	case NODETYPE_MUL:
	case NODETYPE_DIV:
	case NODETYPE_EXP:
		eval(n->expr[0], sym, &v1); coerce(&v1, SCALAR);
		eval(n->expr[0], sym, &v2); coerce(&v2, SCALAR);
		vp->v_type = SCALAR;
		switch (n->type) {
		case NODETYPE_ADD:	
			vp->v_scalar = v1->v_scalar + v2->v_scalar; break;
		case NODETYPE_SUB:
			vp->v_scalar = v1->v_scalar - v2->v_scalar; break;
		case NODETYPE_MUL:
			vp->v_scalar = v1->v_scalar * v2->v_scalar; break;
		case NODETYPE_DIV:
			vp->v_scalar = v1->v_scalar / v2->v_scalar; break;
		case NODETYPE_EXP:
			vp->v_scalar = pow(v1->v_scalar, v2->v_scalar); break;
		}
		break;

	case NODETYPE_INDEX:
		eval(n->expr[0], sym, &v1); coerce(&v1, VECTOR);
		eval(n->expr[1], sym, &v2); coerce(&v2, SCALAR);
		s = eval_index(n, v1.v_vector, v2.v_scalar, vp);
		vector_free(v1.v_vector);
		break;

	case NODETYPE_SUM:
		eval(n->expr[0], sym, &v1); coerce(&v1, VECTOR);
		eval_sum(n, v1.v_vector, vp);
		vector_free(v1.v_vector);
		break;

	case NODETYPE_AVG:
		eval(n->expr[0], sym, &v1); coerce(&v1, VECTOR);
		eval_sum(n, v, vp);
		vp->v_scalar /= (scalar_t)v1.v_vector->len;
		vector_free(v1.v_vector);
		break;

	case NODETYPE_LEN:
		eval(n->expr[0], sym, &v1); coerce(&v1, VECTOR);
		vp->v_ype = SCALAR;
		vp->v_scalar = (scalar_t)v1.v_vector->len;
		vector_free(v1.v_vector);
		break;

	case NODETYPE_IDENT:
		sym_lookup(n->ident, sym, vp);
		/* XXX inc ref count */
		break;

	case NODETYPE_REAL:
		vp->v_type = SCALAR;
		vp->v_scalar = n->real;
		break;

	case NODETYPE_LET:
		y = enter_name(n->ident, sym, n->expr[0]);
		eval(n->expr[1], y, vp);
		sym_leave(y);
		break;

	case NODETYPE_VEC2:
		eval(n->expr[0], sym, vp); coerce(vp, VECTOR);
		eval(n->expr[1], sym, &v1);
		vector_append(vp->v_vector, &v1);
		break;

	case NODETYPE_VEC1:
		eval(n->expr[0], sym, &v1);
		v = new_vector();
		vector_append(v, &v1);
		vp->v_type = VECTOR;
		vp->v_vector = v;
		break;

	case NODETYPE_GEN:
		gen_vector(n, sym, vp);
		break;

	case NODETYPE_RANGE:
		gen_range(n, sym, vp);
		break;

	default:
		eval_error(n, "expected a value");
		/* NOTREACHED */
		return 0;
	}
}

/* Return the expression tree to evaluate */
node_t
eval_lambda(n, sym, idp)
	node_t n;
	sym_t sym;
	ident_t *idp;
{
	node_t n2, n3;
	sym_t y;

	switch (n->type) {
	case NODETYPE_APPLY:
		n2 = eval_lambda(n->expr[0], sym);
		y = enter_name(n2->ident, sym, n->expr[1]);
		n3 = 

		
	case NODETYPE_LAMBDA:
		return n->expr[0];

	default:
		eval_error(n, "expected a function")
		return NULL;
	}
		
}

/* Create a symbol table with a named expression value */
sym_t
enter_name(id, sym, n)
	ident_t id;
	sym_t sym;
	node_t n;
{
	value_t v;
	sym_t y;

	eval_vector(n, sym, &v);
	return sym_enter(&v, id, sym);
}

/* Index into a vector */
void
eval_index(n, v, i, vp)
	node_t n;
	vector_t v;
	scalar_t i;
	value_t *vp;
{
	int idx = SCALAR_ROUND(i);
	if (idx < 0 || idx > v->len)
		eval_error(n, "index out of bounds");
	*vp = v->el[idx];
}

/* Perform a sum over the arguments */
void
eval_sum(n, v, vp)
	node_t n;
	vector_t v;
	value_t *vp;
{
	int i;
	scalar_t val;

	vp->v_type = SCALAR;
	vp->v_scalar = 0;

	for (i = 0; i < v->len; i++) {
		eval_index(n, v, i, &val);
		coerce(&val, SCALAR);
		vp->v_scalar += val.v_scalar;
	}
}

/* Generate a vector */
void
gen_vector(n, sym, vp)
	node_t n;
	sym_t sym;
	value_t *vp;
{
	vector_t v = new_vector();
	value_t src;
	value_t val;
	sym_t y;
	int i;

	eval(n->expr[0], sym, &src); coerce(&src, VECTOR);
	for (i = 0; i < src->v_vector->len; i++) {
		y = sym_enter(&src->v_vector->el[i], n->ident, sym);
		eval(n->expr[1], sym, &val);
		sym_leave(y);
		vector_append(v, &val);
	}

	vp->v_type = VECTOR;
	vp->v_vector = v;
}

void
gen_range(n, sym, vp)
	node_t n;
	sym_t sym;
	value_t *vp;
{
	int i;
	int start, stop;
	value_t startval, stopval;
	vector_t v;

	v = new_vector();
	eval() xxx
	start = SCALAR_ROUND(eval_scalar(n->expr[0], sym));
	stop = SCALAR_ROUND(eval_scalar(n->expr[1], sym));

	if (!(n->flags & RANGE_EXACT_LOWER))
		start++;
	if (!(n->flags & RANGE_EXACT_UPPER))
		stop--;

	for (i = start; i <= stop; i++) {
		vector_append(v, (scalar_t)i);
	}

	return v;
	
}
